"use strict";
var ts = require("typescript");
var path = require("path");
var errors_1 = require("./../../errors");
var utils_1 = require("./../../utils");
var CompilerOptionsResolver = (function () {
    /* istanbul ignore next */ function CompilerOptionsResolver() {
    }
    CompilerOptionsResolver.prototype.getCompilerOptions = function (options) {
        var compilerOptions;
        if (options.compilerOptions != null)
            compilerOptions = utils_1.objectAssign({}, options.compilerOptions || {});
        else if (options.tsConfigFilePath != null)
            compilerOptions = this.getCompilerOptionsFromTsConfig(options.tsConfigFilePath);
        else
            compilerOptions = {};
        return this.getTsCompilerOptionsWithDefaults(compilerOptions);
    };
    CompilerOptionsResolver.prototype.getCompilerOptionsFromTsConfig = function (filePath) {
        this.verifyFileExists(filePath);
        var absoluteFilePath = utils_1.FileUtils.getAbsolutePath(filePath);
        var text = utils_1.FileUtils.readFileSync(absoluteFilePath);
        var result = ts.parseConfigFileTextToJson(absoluteFilePath, text, true);
        if (result.error != null)
            throw new Error(result.error.messageText.toString());
        var settings = ts.convertCompilerOptionsFromJson(result.config.compilerOptions, path.dirname(filePath));
        if (!settings.options)
            throw new Error(this.getErrorMessage(settings.errors));
        return settings.options;
    };
    CompilerOptionsResolver.prototype.getTsCompilerOptionsWithDefaults = function (compilerOptions) {
        function getValue(currentValue, newValue) {
            return (currentValue == null) ? newValue : currentValue;
        }
        var combinedOptions = (compilerOptions || {});
        combinedOptions.allowJs = getValue(combinedOptions.allowJs, true);
        combinedOptions.noLib = getValue(combinedOptions.noLib, false);
        combinedOptions.experimentalDecorators = getValue(combinedOptions.experimentalDecorators, true);
        combinedOptions.suppressExcessPropertyErrors = getValue(combinedOptions.suppressExcessPropertyErrors, true);
        combinedOptions.suppressImplicitAnyIndexErrors = getValue(combinedOptions.suppressImplicitAnyIndexErrors, true);
        combinedOptions.noImplicitAny = getValue(combinedOptions.noImplicitAny, false);
        combinedOptions.target = getValue(combinedOptions.target, ts.ScriptTarget.ES2015);
        combinedOptions.moduleResolution = getValue(combinedOptions.moduleResolution, ts.ModuleResolutionKind.NodeJs);
        combinedOptions.strictNullChecks = getValue(combinedOptions.strictNullChecks, false);
        return combinedOptions;
    };
    CompilerOptionsResolver.prototype.getErrorMessage = function (errors) {
        var message = "";
        errors.forEach(function (err) { return message += err.messageText + "\n"; });
        return message;
    };
    CompilerOptionsResolver.prototype.verifyFileExists = function (filePath) {
        // unfortunately the ts compiler doesn't do things asynchronously so for now we won't either
        if (!utils_1.FileUtils.fileExistsSync(filePath)) {
            throw new errors_1.FileNotFoundError(utils_1.FileUtils.getAbsolutePath(filePath));
        }
    };
    return CompilerOptionsResolver;
}());
exports.CompilerOptionsResolver = CompilerOptionsResolver;

//# sourceMappingURL=CompilerOptionsResolver.js.map
